# python -m sphinx.ext.autosummary.generate C:\ARESYS_PROJ\BioPAL\doc\ICD\index.rst

# python -m sphinx -b=latex -a doc/ICD doc/_build_icd_pdf

# python -m sphinx -b=html -a doc/ICD doc/_build_icd_html

# cd doc/_build_icd_pdf

